package VR.alg

import java.io.PrintWriter

import VR._
import VR.data.DataClass
import VR.prob.ProblemBase
import java.io.File

/**
  * Created by YY on 2016/1/7.
  */
abstract class AlgorithmBase[T<:DataClass](p:ProblemBase[T]) {
  var w:Array[Double]=_
  var prob:ProblemBase[T]=p
  val sampleFreq=5

  case class timeLine ( var last_touch:Int)

  def train(data:T, lambda: Double, sigma : Double, eta: Double, tot_batch:Int, outputFile: Option[String] ): Unit
  def sqr(a:Double)= a*a

  def getTestErr(w:Array[Double], x_dat:X_type, y_dat:Y_type): Double={
    var correct:Int =0
    for (iter<- y_dat.indices){
      if (prob.predict(w,x_dat(iter))==y_dat(iter)) correct+= 1
    }
    1.0- (correct.toDouble/y_dat.length)
  }

  def printToFile(curObj:Double, total_grad:Double, n:Int, f:PrintWriter ) :Double={
    f.println("%.4f".format(total_grad/n)+" %.15f".format(curObj))
    total_grad
  }

  def print_vector(vec:Array[Double], file_name:String):Unit={
    val f_out=new PrintWriter(new File(file_name))
    for (i<- vec) {
        f_out.write(i+" ")
    }
    f_out.write("\n")
    f_out.close()
  }

  def print_sparse_vector(vec:Array[Sparse_type], file_name:String):Unit={
    val f_out=new PrintWriter(new File(file_name))
    for (i<- vec) {
      f_out.write(i._2+" ")
    }
    f_out.write("\n")
    f_out.close()
  }

  def update_w(cur_val:Double, tl:timeLine, curT: Int, save: Array[Double], sp:Int, proximalThres:Double, shrink: Double, normal_grad:Double ):Double={
    var totR:Int = curT-1 - tl.last_touch
    tl.last_touch=curT-1
    if (totR<=0)
      return cur_val
    if (proximalThres<=1e-10){
      if (shrink>=1.0-1e-20){
        save(sp) += totR * cur_val
        return cur_val
      }
      val t:Double=shrink
      val progSum:Double= (1.0-math.pow(t,totR))/(1.0-t)
      save(sp) += cur_val*t*progSum - normal_grad*t / (1.0-t) * (totR - t*progSum)
      return cur_val*math.pow(t,totR)-normal_grad*t*progSum
    }
    if (cur_val >= 0)
    {
      if (normal_grad > 0)
      {
        //n+p>=0
        val stepAppr:Int = (cur_val / (normal_grad + proximalThres)).toInt
        if (stepAppr >= totR)
        {
          save(sp) += (cur_val - (normal_grad + proximalThres)*(1.0 + totR) / 2.0)*totR
          return cur_val - (normal_grad + proximalThres)*totR
        }
        totR -= stepAppr
        save(sp) += (cur_val - (normal_grad + proximalThres)*(1.0 + stepAppr) / 2.0)*stepAppr
        var val_ =cur_val
        val_ -= (normal_grad + proximalThres)*stepAppr

        //never escape
        if (normal_grad <= proximalThres)
          return 0
        totR-=1
        val_ -= normal_grad
        val_ += proximalThres
        if (val_ >= 0)
          val_ = 0
        save(sp) += (val_ - (normal_grad - proximalThres)*(0.0 + totR) / 2.0)*(totR + 1)
        val_ - (normal_grad - proximalThres)*totR
      }
      else
      //below: tl.normal_grad<=0
      {
        if (normal_grad + proximalThres <= 1e-8)
        {
          save(sp) += (cur_val - (normal_grad + proximalThres)*(1.0 + totR) / 2.0)*totR
          //-|n|+p<=0, go more and more positive
          return cur_val - (normal_grad + proximalThres)*totR
        }
        //below:  -|n|+p>0, pull close to 0
        val stepAppr:Int = (cur_val / (normal_grad + proximalThres)).toInt
        if (stepAppr >= totR)
        {
          save(sp) += (cur_val - (normal_grad + proximalThres)*(1.0 + totR) / 2.0)*totR
          return cur_val - (normal_grad + proximalThres)*totR
        }
        save(sp) += (cur_val - (normal_grad + proximalThres)*(1.0 + stepAppr) / 2.0)*stepAppr
        //else,never escape
        0
      }
    }
    else
    //val<0
    {
      if (normal_grad < 0)
      {
        // -|n|-p=n-p<0, which is the progress
        // -val/ |n-p|= -val / (p-n)
        val stepAppr:Int = (-cur_val / (-normal_grad + proximalThres)).toInt
        if (stepAppr >= totR)
        {
          save(sp) += (cur_val - (normal_grad - proximalThres)*(1.0 + totR) / 2.0)*totR
          return cur_val - (normal_grad - proximalThres)*totR
        }
        totR -= stepAppr
        save(sp) += (cur_val - (normal_grad - proximalThres)*(1.0 + stepAppr) / 2.0)*stepAppr
        var val_ = cur_val
        val_ -= (normal_grad - proximalThres)*stepAppr
        if (-normal_grad <= proximalThres)
        //never escape
          return 0
        //below n+p<0
        totR-=1
        val_ -= normal_grad
        val_ -= proximalThres
        if (val_ <= 0)
          val_ = 0
        save(sp) += (val_ - (normal_grad + proximalThres)*(0.0 + totR) / 2.0)*(totR + 1)
        val_ - (normal_grad + proximalThres)*totR
      }
      else
      //below: tl.normal_grad>=0
      {
        if (normal_grad - proximalThres >= -1e-8)
        {
          save(sp) += (cur_val - (normal_grad - proximalThres)*(1.0 + totR) / 2.0)*totR
          // n - p >=0, go more and more negative
          return cur_val - (normal_grad - proximalThres)*totR
        }
        //below:  n-p<0, pull close to 0
        val stepAppr:Int = (-cur_val / (-normal_grad + proximalThres)).toInt
        if (stepAppr >= totR)
        {
          save(sp) += (cur_val - (normal_grad - proximalThres)*(1.0 + totR) / 2.0)*totR
          return cur_val - (normal_grad - proximalThres)*totR
        }
        save(sp) += (cur_val - (normal_grad - proximalThres)*(1.0 + stepAppr) / 2.0)*stepAppr
        //else,never escape
        0
      }
    }
  }
}
