package VR.alg

import java.io.{File, PrintWriter}

import VR.data.DataClass
import VR. Sparse_type
import VR.prob.ProblemBase

import scala.util.Random

/**
  * Created by YY on 2016/1/7.
  */
class SVRG[T<:DataClass](p:ProblemBase[T]) extends AlgorithmBase(p){

  case class ClusterInfo (var lastVisitedRow:Int,var ClusterSize: Int, delta:Array[Double])

  def extra_update(w:Array[Double], save:Array[Double], m:Int):Unit={}

  def get_m(n:Int):Int= n*2
  def update_m(m:Int):Int = m
  def update_wtilde(wtilde:Array[Double], w:Array[Double]):Unit={
    for (j<-w.indices)
      wtilde(j)=w(j)
  }
  def stop_this_epoch(t:Int, m:Int):Boolean= t>=m
  def train_init(m:Int):Unit={}

  var cBlock:Array[ClusterInfo]=_
  var sCluster:Int=0

  def update_grad_info(t:Int,t1:Array[Double], t2:Array[Double]):Unit= {}

  def Cluster_preprocess(belongTo:Array[Int], n:Int, d:Int):Unit={
    for (i<-belongTo.indices)
      if (sCluster <= belongTo(i))
        sCluster=belongTo(i)+1
    cBlock=new Array[ClusterInfo](sCluster)
    for (i<-Range(0,sCluster)) cBlock(i)=new ClusterInfo(-1, 0, new Array[Double](d))
    for (i<-belongTo.indices) cBlock(belongTo(i)).ClusterSize += 1
  }

  override def train(data:T,  lambda: Double, sigma : Double, eta: Double,  tot_batch:Int,outputFile: Option[String]): Unit = {
    val (x_dat,_,n,d)=data.get_data()
    val belongTo=data.belongTo
    Cluster_preprocess(belongTo,n,d)
    w=new Array[Double](d)

    val wtilde:Array[Double]=new Array[Double](d)
    val utilde:Array[Double]=new Array[Double](d)
    val save: Array[Double]=new Array[Double](d)
    val metBefore: Array[Int]= new Array[Int](d)
    val thres:Double =math.abs(eta*sigma)
    var m:Int=get_m(n)

    train_init(m)

    val timeline:Array[timeLine]=new Array[timeLine](d)
    for (i<- timeline.indices) timeline(i)=new timeLine(-1)
    val fout:Option[PrintWriter]=outputFile match {
      case None => None
      case Some(st)=>Some(new PrintWriter(new File(st)))}
    var totGrad:Double=0
    var lastPrint:Double=0
    var get_out:Boolean=false
    for (i<-wtilde.indices) wtilde(i)=w(i)

    val r_seed=new Random(1)
    while (!get_out)
    {
      for (i<-save.indices) save(i)=0

      //below, compute the actual full gradient, utilde is not sparse.
      prob.fullGrad(data,wtilde, utilde)
      totGrad += n
      for (curBlock<-cBlock){
        for (i<-curBlock.delta.indices)
          curBlock.delta(i)=0
        curBlock.lastVisitedRow = -1
      }
      for (i<-metBefore.indices) metBefore(i)=0

      for (i <- timeline.indices) timeline(i).last_touch = -1

      var t:Int=0
      while (! stop_this_epoch(t,m)){
        val randomRow:Int=r_seed.nextInt(n)
        val i:Int=belongTo(randomRow)
        val Bi:ClusterInfo=cBlock(i)
        val vR:Int=Bi.lastVisitedRow

        // all coordinates related to j are updated.
        for (j<- x_dat(randomRow))
          w(j._1) = update_w(w(j._1),timeline(j._1), t, save, j._1, thres, 1.0-eta*lambda, eta*utilde(j._1))
        // all coordinates related to j_i are also updated.
        if (vR >= 0)
          for (j<-x_dat(vR))
            w(j._1) = update_w(w(j._1),timeline(j._1), t, save, j._1, thres, 1.0-eta*lambda, eta*utilde(j._1))

        //compute g1 and g2 here.
        val t1:Array[Double]=prob.gradient_cal(w,data, randomRow)
        val t2:Array[Double]=prob.gradient_cal(wtilde,data,randomRow)
        update_grad_info(t,t1,t2)

        // - eta(utilde + \nabla_ f -\nabla_i f')
        for (j<-x_dat(randomRow).indices){
          // w - eta( utilde + g1-g2)
          val where=x_dat(randomRow)(j)._1
          w(where) -= eta*((t1(j)-t2(j)) + utilde(where))
          metBefore(where)=t
        }
        if (vR >= 0)
        {
          // - eta(- delta) only if vR>=0
          // also remove the corresponding part in utilde
          for (j<-x_dat(vR).indices){
            val curS:Sparse_type=x_dat(vR)(j)
            val Bi_d:Double=Bi.delta(j)
            // w - eta * old delta
            if (metBefore(curS._1)==t)
              w(curS._1) -= eta* (-Bi_d)
            else
              w(curS._1) -= eta* (utilde(curS._1) - Bi_d)
            // utilde + (old delta)* size/n
            utilde(curS._1) -= Bi_d * Bi.ClusterSize /n
          }
          //proximal updates
          for (j<-x_dat(vR)){
            if (sigma>1e-10)
            {
              if (w(j._1) > thres)
                w(j._1) -= thres
              else if (w(j._1) < -thres)
                w(j._1) += thres
              else w(j._1)=0
            }
            else
              w(j._1) *= 1.0-eta*lambda
            timeline(j._1).last_touch=t
            save(j._1) += w(j._1)
          }
        }
        for (j<-x_dat(randomRow))
        // if this coordinate is visited before, skip this one.
          if (timeline(j._1).last_touch !=t)
          {
            if (sigma>1e-10)
            {
              if (w(j._1)>thres)
                w(j._1) -= thres
              else if (w(j._1)< -thres)
                w(j._1) += thres
              else
                w(j._1)=0
            }
            else
              w(j._1) *= 1.0- eta*lambda
            timeline(j._1).last_touch=t
            save(j._1) += w(j._1)
          }

        //update the lastVisitedRow
        Bi.lastVisitedRow = randomRow

        //update delta as g1-g2
        // update utilde as well.
        for (j<-x_dat(randomRow).indices){
          val curS:Sparse_type=x_dat(randomRow)(j)
          Bi.delta(j) = t1(j)-t2(j)
          utilde(curS._1) += Bi.delta(j) * Bi.ClusterSize /n
        }
        totGrad+=1

        if (totGrad > lastPrint + n / sampleFreq)
        {
          for (j<-w.indices)
            w(j)=update_w(w(j),timeline(j), t+1, save, j, eta*sigma, 1.0-eta*lambda, eta*utilde(j))
          val curObj:Double=prob.full_obj_cal(data,w,lambda,sigma)
          fout match {
            case Some(a:PrintWriter)=>
              lastPrint=printToFile(curObj,totGrad,n,a)
            case None =>
          }
        }

        /*
        for (j<-w.indices){
          w(j)=update_w(w(j),timeline(j), t+1, save, j, eta*sigma, 1.0-eta*lambda, eta*utilde(j))
          timeline(j).last_touch=t
        }
        */

        t=t+1
      }
      for (j<-w.indices){
        w(j)=update_w(w(j),timeline(j), t, save, j, eta*sigma, 1.0-eta*lambda, eta*utilde(j))
        wtilde(j)=save(j)/t
      }
      update_wtilde(wtilde,w)
      extra_update(w,save,t)
      m= update_m(m)
      val curObj:Double=prob.full_obj_cal(data,w,lambda,sigma)
      println("%.10f  eta=%.5f  w0=%.5f  w1=%.5f  w2=%.5f".format(curObj,eta,w(0), w(1), w(2)))

      if (totGrad/tot_batch>=n)
        get_out=true
    }

    fout match {
      case Some(a:PrintWriter) => a.close()
      case None =>
    }
  }

}
