package VR.my_util

import java.io.{File,PrintWriter}

import VR._
import VR.alg._
import VR.data.{DataProcessor, DataClass}
import VR.prob.{LinearRegression, LogisticRegression, ProblemBase}

import scala.util.Random

/**
  * Created by YY on 2016/1/23.
  */
object ERM_worker extends Worker {
  def cross_validation(data:DataClass, target:String, eta:Double, batch:Int, prob:ProblemBase[DataClass]):Unit ={
    val (x_dat,y_dat,n,d)=data.get_data()
    val r_seed=new Random(1)
    val split_x=Array(new X_type(n),new X_type(n))
    val split_y=Array(new Y_type(n),new Y_type(n))
    val label=Array[Int](n)
    val fold=10
    for (i<-y_dat.indices)
      label(i)=r_seed.nextInt(fold)
    val parameter_list:List[Double]=List(3,1,3e-1, 1e-1, 3e-2, 1e-2, 3e-3, 1e-3, 3e-4, 1e-4, 3e-5, 1e-5)
    var best_error:Double=1.0
    var which_para:Double= -1
    for (para <- parameter_list){
      println("Trying %f\n".format(para))
      var cumulative_error:Double=0
      for (curfold <- Range(0,fold)){
        val count=Array(0,0)
        for (i <- y_dat.indices){
          val belong= if (label(i)!=curfold) 0 else 1
          split_y(belong)(count(belong))=y_dat(i)
          split_x(belong)(count(belong))=x_dat(i)
          count(belong)+=1
        }
        val alg_uniVR:UniVR[DataClass]=new UniVR[DataClass](prob)
        if (target.equals("sigma"))
          alg_uniVR.train(new DataClass(split_x(0),split_y(0),split_x(0).length, d),0,para,eta,batch,None)
        else
          alg_uniVR.train(new DataClass(split_x(0),split_y(0),split_x(0).length, d),para,0,eta,batch,None)
        cumulative_error+= alg_uniVR.getTestErr(alg_uniVR.w,split_x(1),split_y(1))
      }
      cumulative_error /= fold
      println("Cumulative Error is %.4f\n".format(cumulative_error))
      if (best_error> cumulative_error){
        best_error = cumulative_error
        which_para=para
      }
    }
    println("Done Cross Validation with UniVR! The best parameter of %s is %g\n".format(target,which_para))
  }
  def do_work(options:OptionMap):Unit={
    print_options(options)

    var train_data:DataClass=
      if (options.contains('cluster))
        new DataClass(options('datafolder)+""+options('dataname)+".dat", options('datafolder)+""+options('dataname)+"-helper.txt", options.contains('logistic), options('cluster).toString)
      else
        new DataClass(options('datafolder)+""+options('dataname)+".dat", options('datafolder)+""+options('dataname)+"-helper.txt", options.contains('logistic))

    val prob:ProblemBase[DataClass]= if (options.contains('logistic)) new LogisticRegression() else new LinearRegression()

    if (options.contains('cross)){
      cross_validation(train_data, options('cross).toString, options('eta).toString.toDouble, options('batch).toString.toInt, prob)
    }
    else {
      var test_dat:Option[DataClass]=None
      var test_output_name:String=""
      if (options.contains('split)){
        test_output_name="split"   //split
        val (a,b)=DataProcessor.random_split(train_data)
        train_data=a
        test_dat=Some(b)
      } else
      if (options.contains('testdataname)){
        test_output_name="test" //test
        test_dat=Some(new DataClass(DataProcessor.dataReader(options('datafolder)+""+options('testdataname)+".dat",
          options('datafolder)+""+options('testdataname)+"-helper.txt", options.contains('logistic))))
      }

      val outputFolder= if (options.contains('lambda)) options('dataname)+"-ridge-"+options('lambda)+"\\"  //a9a-ridge-1e-5\
      else if (options.contains('logistic)) options('dataname)+"-logistic-l1-"+options('sigma)+"\\"  //a9a-logistic-l1-1e-5\
      else options('dataname)+"-lasso-"+options('sigma)+"\\"   //a9a-lasso-1e-5\

      val success = new File(outputFolder).mkdirs()
      if (!success) println("Failed to make a new folder. It may already exist.")

      val lambda:Double=if (options.contains('lambda)) options('lambda).toString.toDouble else 0
      val reg_lambda:Double=if (options.contains('reglambda)) options('reglambda).toString.toDouble else 0
      val sigma:Double= if (options.contains('sigma)) options('sigma).toString.toDouble else 0
      val eta:Double = options('eta).toString.toDouble
      val tot_batch:Int=options('batch).toString.toInt+2
      val scaled_value:Int=1000000
      val outputFileName:String= if (options.contains('cluster)){
        val cluster_name:String=options('cluster).toString  //a9a-0-2800.dat
        val reduced_name:String=cluster_name.substring(0,cluster_name.length-4)  //a9a-0-2800
        reduced_name+"-Local"+options('algorithm).toString+"-"+(eta*scaled_value).toInt.toString  //a9a-0-2800-LocalSVRG-30000
      } else {
        if (options('algorithm).toString.toUpperCase=="SDCA")
          options('dataname)+"-"+options('algorithm).toString+"-r"+reg_lambda.toString //a9a-SDCA-r1e-5
        else
          options('dataname)+"-"+options('algorithm).toString+"-"+(eta*scaled_value).toInt.toString  //a9a-SVRG-30000
      }
      val cur_alg:AlgorithmBase[DataClass]=
        options('algorithm) match
        {
          case "SVRG" => new SVRG(prob)
          case "UNIVR" => new UniVR(prob)
          case "SAGA" =>  new SAGA(prob)
          case "SAG" =>  new SAG(prob)
          case "SDCA" => new SDCA(prob, reg_lambda)
          case _ => println("ERROR! Unknown algorithm name. Using SAGA instead. \n"); assert(false); new SAGA(prob)
        }
      cur_alg.train(train_data,lambda,sigma,eta,tot_batch, Some(outputFolder+""+outputFileName)) //a9a-0-1e-5\ a9a-0-2800-LocalSVRG-30000

      test_dat match {
        case Some(t_dat) =>
          val testErr:Double=cur_alg.getTestErr(cur_alg.w,t_dat.x, t_dat.y)
          test_output_name =outputFileName+"-"+test_output_name + "-error"  // a9a-0-2800-LocalSVRG-30000-split/test-error
        val f_test_out=new PrintWriter(new File(outputFolder+test_output_name))
          f_test_out.write(testErr.toString)
          f_test_out.close()
        case None =>
      }
    }
  }

}
